#include <stdio.h>
#include <string.h>
#include <math.h>
#include "fitsio.h"
#include "pil.h"
#include "headas.h"
#include "headas_error.h"
#include "ftstat.h"

#define MAXMSG 256


/*
HISTORY
-------
  Version 1.0 written by William Pence and Ziqin Pan, NASA/GSFC, October 2002
  Version 1.1 Added median output & quickselect routine; Bryan Irby, Feb 2009
*/

#define TOOLSUB ftstat
/* headas_main() requires that TOOLSUB be defined first */
#include "headas_main.c"

/* Function Prototypes */
int ftstat (void);
int ftstat_getpar (char *infile, char *outfile, int *centroid, 
                int *clip, double *nsigma,int *maxiter);
int ftstat_work (char *infile, char *outfile, int centroid, int clip, double nsigma,
                 int maxiter); 
int cal_tbl_stat(hduInfo * info, int centroid, int clip, 
                double nsigma, int maxiter,int *status);
int cal_img_stat(fitsfile* infptr, hduInfo * info, int centroid, int clip,
                double nsigma, int maxiter, int *status);
int cal_col_stat(Column * column, int *status);
int sigma_clip(Column *column, double nsigma, int maxiter, int *status);
int print_stat(FILE* outfptr, char* infile, 
                hduInfo info, int centroid, int clip);
int read_hdu_data(fitsfile * infptr, hduInfo *info, int * status);
int ftstat_putpar(hduInfo * info, int centroid, int* status);
void ftstat_free(hduInfo * info);
static double ftstat_median_double(double arr[], int n);
/*---------------------------------------------------------------------------*/
int ftstat (void)
{
/*  Compute statistics for 2D images and table columns */

    char infile[PIL_LINESIZE], outfile[PIL_LINESIZE];
    int centroid, clip, maxiter, status;
    double nsigma;
    static char taskname[80] = "ftstat";
    static char version[8] = "1.10";

    /* Register taskname and version. */
    set_toolname(taskname);
    set_toolversion(version);

    /* Get input parameters */
    status = ftstat_getpar(infile, outfile, &centroid, &clip, &nsigma, 
                 &maxiter);

    /* Call work function to compute statistics for 2D images 
       and table columns */
    if (!status)
        status = ftstat_work(infile, outfile, centroid, clip, nsigma,
                   maxiter);

    return(status);
}
/*---------------------------------------------------------------------------*/
int ftstat_getpar(
    char *infile,    /* O - Input file name */ 
    char *outfile,   /* O - optional output file name */
    int *centroid,   /* O - Whether to compute image centroid */
    int *clip,       /* O - whether to iteratively reject N*sigma
                            outliers and recompute the statistics */
    double *nsigma,  /* O - if clip = yes, then this is the sigma
                            rejection factor */
    int *maxiter)    /* O - if clip = yes this is the maximum number
                            of iterations */ 

/*  read input parameters for the ftstat task from the .par file */
{
    int status;
    char msg[MAXMSG];

    if ((status = PILGetFname("infile", infile))) {
      sprintf(msg, "Error reading the 'infile' parameter.");
      HD_ERROR_THROW(msg,status);
    }

    else if ((status = PILGetFname("outfile", outfile))) {
      sprintf(msg, "Error reading the 'outfile' parameter.");
      HD_ERROR_THROW(msg,status);
    }

    else if ((status = PILGetBool("centroid", centroid))) {
      sprintf(msg, "Error reading the 'centroid' parameter.");
      HD_ERROR_THROW(msg,status);
    }


    else if ((status = PILGetBool("clip", clip))) {
      sprintf(msg, "Error reading the 'clip' parameter.");
      HD_ERROR_THROW(msg,status);
    }


    else if ((status = PILGetReal("nsigma", nsigma))) {
      sprintf(msg, "Error reading the 'nsigma' parameter.");
      HD_ERROR_THROW(msg,status);
    }

    else if ((status = PILGetInt("maxiter", maxiter))) {
      sprintf(msg, "Error reading the 'maxiter' parameter.");
      HD_ERROR_THROW(msg,status);
    }


    return(status);
}
/*---------------------------------------------------------------------------*/
int ftstat_work(
    char *infile,   /* I - Input file name */ 
    char *outfile,  /* I - Output file name */ 
    int centroid,   /* I - Whether to compute image centroid */
    int clip,       /* I - whether to iteratively reject N*sigma
                            outliers and recompute the statistics */
    double nsigma,     /* I - if clip = yes, then this is the sigma
                            rejection factor */
    int maxiter)    /* I - if clip = yes this is the maximum number
                            of iterations */ 
{
    FILE *stream;
    fitsfile *infptr = 0;
    int status = 0, len;
    char *cptr;
    char msg[MAXMSG];
    hduInfo info;

    info.column = 0;
    
    /* strip leading and trailing spaces off of 'outfile' */
    cptr = outfile;
    while (*cptr == ' ')cptr++;
    len = strlen(cptr);
    while (len > 0 && *(cptr+len-1) == ' ') {
        len--;
        *(cptr+len) = '\0';
    }

    /* determine output stream (stdout or to a file) */
    if (*cptr == '\0' || !strcmp(cptr, "-") || 
        !strcmp(cptr, "stdout") || !strcmp(cptr, "STDOUT") ) {
        stream = stdout;
    } else {
        if (*cptr == '!') {  /* clobber the output file */
            cptr++;
            remove(cptr);
        } else {
            headas_clobberfile(cptr);  /* delete existing file if clobber=YES */
        }

        stream = fopen(cptr, "r"); /* does file already exist? */

        if (stream)
        {
            fclose(stream);         /* close file and exit with error */
            sprintf(msg,"Error: Output file already exists:\n %s",cptr);
            status = FILE_NOT_CREATED;
            HD_ERROR_THROW(msg,status);
            return(status); 
        }

        if (!(stream = fopen(cptr, "w"))) {
            sprintf(msg,"Error creating output file:\n %s",cptr);
            status = FILE_NOT_CREATED;
            HD_ERROR_THROW(msg,status);
            return(status); 
        }
    }

    /* Open the input file, and move to first 'interesting' HDU */
    if ( fits_open_data(&infptr, infile, READONLY, &status)) goto cleanup;
    headas_chat(5,"Opened the input file:\n %s\n",infile);

    /* get the current HDU data*/
    if ( read_hdu_data(infptr,&info,&status)) goto cleanup;
    headas_chat(5,"Read data from current HDU\n");
   
    if( info.hdutype == IMAGE_HDU)
        cal_img_stat(infptr,&info,centroid,clip,nsigma,maxiter,&status);
    else if( info.hdutype == ASCII_TBL || info.hdutype == BINARY_TBL )
        cal_tbl_stat(&info,0,clip,nsigma,maxiter,&status);

    print_stat(stream,infile,info,centroid,clip);

    ftstat_putpar(&info,centroid,&status);


cleanup:

    if (infptr)  fits_close_file(infptr,  &status);
    if (status) fits_report_error(stderr, status);
    if (info.column)  ftstat_free(&info);
    return(status);
}

int cal_tbl_stat(hduInfo * info, int centroid, int clip, 
      double nsigma, int maxiter,int *status)
{
    int j;

    for (j=0; j<info->ncols; j++) 
    {
         cal_col_stat(info->column+j,status);

         if(clip && !centroid) sigma_clip(info->column+j,nsigma,maxiter,status);
    }

    return *status;
}

int cal_img_stat(fitsfile* infptr, hduInfo * info, int centroid, int clip,
       double nsigma, int maxiter, int *status)
{
    long i;
    long x, y;
    double sumx,sumy,sumx2,sumy2;
    double xrefval=0., yrefval=0., xrefpix=0.;
    double yrefpix=0., xinc=0., yinc=0., rot=0.;
    char coordtype[20];

    info->column[0].wcflg =TRUE;
    if(fits_read_img_coord(infptr,&xrefval,&yrefval,&xrefpix,
           &yrefpix,&xinc,&yinc,&rot,coordtype,status))  
    {
         info->column[0].wcflg=FALSE;
         *status =0;
    }

    
    cal_tbl_stat(info,centroid,clip,nsigma,maxiter,status);
    info->column[0].ymin = info->column[0].xmin/info->naxes[0]+1;
    info->column[0].xmin = info->column[0].xmin%info->naxes[0]+1;
    info->column[0].ymax = info->column[0].xmax/info->naxes[0]+1;
    info->column[0].xmax = info->column[0].xmax%info->naxes[0]+1;


    if(centroid) 
    {
        sumx=0.;
        sumy=0.;
        sumx2=0.;
        sumy2=0.;
        for (i=0; i<info->column[0].total; i++)
        {
           if( !info->column[0].nulflag[i])
           {
              x =i%info->naxes[0]+1;
              y =i/info->naxes[0]+1;
              sumx +=x*info->column[0].dvalue[i];
              sumy +=y*info->column[0].dvalue[i];
              sumx2 +=x*x*info->column[0].dvalue[i];
              sumy2 +=y*y*info->column[0].dvalue[i];
           }
        }

        if( info->column[0].sum > 0.) 
        {
           info->column[0].cx=sumx/info->column[0].sum;
           info->column[0].cy=sumy/info->column[0].sum;
           info->column[0].dx =sqrt(sumx2/info->column[0].sum -
                                   info->column[0].cx*info->column[0].cx);
           info->column[0].dy =sqrt(sumy2/info->column[0].sum -
                                   info->column[0].cy*info->column[0].cy);
        }

    }

    if(info->column[0].wcflg) 
    {

    if(fits_pix_to_world(info->column[0].xmin, info->column[0].ymin,
       xrefval, yrefval,xrefpix, yrefpix, xinc, yinc,rot,
       coordtype, &info->column[0].rxmin,&info->column[0].rymin,status)) 
    {
       info->column[0].wcflg = FALSE;
       *status =0;
       return *status;
    }

    if(fits_pix_to_world(info->column[0].xmax, info->column[0].ymax,
       xrefval, yrefval,xrefpix, yrefpix, xinc, yinc,rot,
       coordtype, &info->column[0].rxmax,&info->column[0].rymax,status))
    {
       info->column[0].wcflg = FALSE;
       *status =0;
       return *status;
    }

    if(fits_pix_to_world(info->column[0].cx, info->column[0].cy,
       xrefval, yrefval,xrefpix, yrefpix, xinc, yinc,rot,
       coordtype, &info->column[0].rcx,&info->column[0].rcy,status)) 
    {
       info->column[0].wcflg = FALSE;
       *status =0;
       return *status;
    }

    if(fits_pix_to_world(info->column[0].dx+info->column[0].cx, 
                         info->column[0].dy+info->column[0].cy,
       xrefval, yrefval,xrefpix, yrefpix, xinc, yinc,rot,
       coordtype, &info->column[0].rdx,&info->column[0].rdy,status)) 
    {
       info->column[0].wcflg = FALSE;
       *status =0;
       return *status;
    }
     info->column[0].rdx -= info->column[0].rcx;
     info->column[0].rdy -= info->column[0].rcy;
    if(info->column[0].rdx <0) info->column[0].rdx=-info->column[0].rdx;
    if(info->column[0].rdy <0) info->column[0].rdy=-info->column[0].rdy;
    }

    return *status;

}

int read_hdu_data(fitsfile * infptr, hduInfo *info, int * status)
{
    int j;
    int anyf;
    long anyl;
    char str[20];
    char msg[MAXMSG];

    /* Initialize naxes[] */
    info->naxes[0]=0;
    info->naxes[1]=0;

    if (fits_get_hdu_type(infptr, &(info->hdutype), status)) return(*status);

    if ( info->hdutype == ASCII_TBL || info->hdutype == BINARY_TBL )
    {
          if(fits_get_num_cols(infptr,&(info->ncols),status)) return(*status);
          if(fits_get_num_rows(infptr,&(info->nrows),status)) return(*status);
          info->column = (Column *) calloc(info->ncols,sizeof(Column));

          for (j=0; j<info->ncols; j++)
          {
              sprintf(str,"%d",j+1);
              if(fits_get_colname(infptr,
                  1,str,info->column[j].name,&anyf,status)) return(*status);
              if(fits_get_coltype(infptr,j+1,&(info->column[j].dtype),
                  &anyl,&anyl,status)) return(*status);
              sprintf(str,"%s%d","TUNIT",j+1);
              if(fits_read_key(infptr,TSTRING,str,info->column[j].unit,
                  NULL,status)) 
              {  
                   strcpy(info->column[j].unit," \0");
                   *status =0;
              }

              (info->column[j]).dvalue =(double*) calloc(info->nrows,sizeof(double));
              (info->column[j]).nulflag =(char*) calloc(info->nrows,sizeof(char));
              info->column[j].total = info->nrows;

	    if(info->column[j].dtype == TBIT ||
	       info->column[j].dtype == TBYTE ||
	/*       info->column[j].dtype == TLOGICAL ||  */
	       info->column[j].dtype == TUSHORT ||
	       info->column[j].dtype == TSHORT ||
	       info->column[j].dtype == TUINT ||
	       info->column[j].dtype == TINT ||
	       info->column[j].dtype == TULONG ||
	       info->column[j].dtype == TLONG ||
	       info->column[j].dtype == TFLOAT ||
	       info->column[j].dtype == TDOUBLE )
	    {
              if(fits_read_colnull_dbl(infptr,j+1,1,1,info->nrows,
                (info->column[j]).dvalue,(info->column[j]).nulflag,&anyf,status))
                 return(*status);
	    }

          }
    }
    else if (info->hdutype == IMAGE_HDU)
    {
          if(fits_read_key(infptr,TSTRING,"XTENSION",info->extname,
                  NULL,status)==0) 
          {
               
             if(fits_read_key(infptr,TSTRING,"EXTNAME",info->extname,
                  NULL,status)) 
             {
               strcpy(info->extname," \0");
               *status=0;
             }
          }
          else 
          {
               strcpy(info->extname,"PRIMARY\0");
               *status=0;
          }

          if(fits_get_img_dim(infptr,&(info->naxis),status)) return(*status);
          if(info->naxis > 2 || info->naxis < 1) 
          {
              sprintf(msg,"NAXIS = %d, ftstat can be only applied to 1-D or 2-D image", 
              info->naxis);
              info->column = 0;
              *status =108;
              HD_ERROR_THROW(msg,*status);
              return *status;
          }
          if(fits_get_img_size(infptr,info->naxis,info->naxes,status)) return(*status);
          info->ncols=1;
          if(info->naxes[1] == 0) /* 1-D array */
          {
            info->nrows=info->naxes[0];
          }
          else
          {
            info->nrows=info->naxes[0]*info->naxes[1];
          }
          info->column = (Column *) calloc(info->ncols,sizeof(Column));
          info->column[0].total =info->nrows;
          info->column[0].dtype =TDOUBLE;
          info->column[0].dvalue =(double*) calloc(info->nrows,sizeof(double));
          info->column[0].nulflag =(char*) calloc(info->nrows,sizeof(char));
          if(fits_read_imgnull(infptr,TDOUBLE,1,info->column[0].total,
                info->column[0].dvalue, info->column[0].nulflag, &anyf,status))
                return *status;
    }
          
    return(*status);
}

int cal_col_stat(Column * column, int *status)
{
    int i,j,k;
    int first;
       column->sum =0.;
       column->sumsq =0.;
       column->mean =0.;
       column->median =0.;
       column->max =0.;
       column->min =0.;
       column->std =0.;
       column->good =0;
       column->null =0;
       column->ymin =1;
       column->ymax =1;
       column->cx =0;
       column->cy =0;
       column->dx =0;
       column->dy =0;

    if(column->dtype == TBIT ||
       column->dtype == TBYTE ||
    /*   column->dtype == TLOGICAL || */
       column->dtype == TUSHORT ||
       column->dtype == TSHORT ||
       column->dtype == TUINT ||
       column->dtype == TINT ||
       column->dtype == TULONG ||
       column->dtype == TLONG ||
       column->dtype == TFLOAT ||
       column->dtype == TDOUBLE )
    {
      
       first =TRUE;
        
       for ( i=0; i<column->total; i++)
       {
          if(column->nulflag[i] ==0) 
          {
            (column->good)++;
            column->sum +=column->dvalue[i];
            column->sumsq +=column->dvalue[i] *
                      column->dvalue[i] ;
            if(first)
            {
                  column->max = column->dvalue[i];
                  column->min = column->dvalue[i];
                  column->xmax =i;
                  column->xmin =i;
                  first = FALSE;
            }

            if( column->max < column->dvalue[i] )
            {
              
                  column->max =column->dvalue[i];
                  column->xmax =i;
            }
                 
            if( column->min > column->dvalue[i] )
            {
                  column->min =column->dvalue[i];
                  column->xmin =i;
            }
         } 
         else (column->null)++;
          
       }

       if(column->good > 0) 
       {
           column->mean =column->sum/column->good;
           column->std =sqrt(column->sumsq/column->good 
                - column->mean*column->mean);

           /* Remove NULLs before computing median: */
           if (column->null > 0)
           {
               k=0;
               for (j=0; j<column->total; j++)
               {
                   /* Shift good values into the lower column->good elements: */
                   if (column->nulflag[j]) k++;
                   else column->dvalue[j-k] = column->dvalue[j];
               }
           }
           column->median = ftstat_median_double(column->dvalue,column->good);

       }
    }

    return *status;
}

int sigma_clip(Column *column, double nsigma, int maxiter, int *status) 
{
    int i;

    if(maxiter ==0 ) return (*status);
    if(column->converge) return (*status);                                             

    column->converge =TRUE;
    for (i=0;i<column->total; i++)
    {
        if(column->nulflag[i] == FALSE) 
        {
           if( column->dvalue[i] >  column->mean + nsigma* column->std ||
               column->dvalue[i] <  column->mean - nsigma* column->std )
           {
               column->converge = FALSE;
               column->nulflag[i] =TRUE;
           }
        }
         
    }

    cal_col_stat(column,status);

    sigma_clip(column,nsigma, maxiter-1, status);

    return (*status);
}


        

int print_stat(FILE* outfptr, char* infile, hduInfo info, int centroid, int clip) {


 int i;

    fprintf(outfptr,"====== statistics for %s ======\n",infile);
 for (i=0; i<info.ncols; i++) {

    fprintf(outfptr,"\n");
    if(info.hdutype == IMAGE_HDU)
    {
       if(strcmp(info.extname," ") !=0) {
          fprintf(outfptr,"           extname:  %s\n",info.extname);
       }
       fprintf(outfptr,"     minimum value:  %f\n", info.column[i].min);
       fprintf(outfptr,"       pixel coord:  (%ld,%ld)\n", info.column[i].xmin,info.column[i].ymin);

       if(info.column[i].wcflg) 
       {
          fprintf(outfptr,"       world coord:  (%g,%g)\n",
             info.column[i].rxmin,info.column[i].rymin);
       }

       fprintf(outfptr,"     maximum value:  %g\n", info.column[i].max);
       fprintf(outfptr,"       pixel coord:  (%ld,%ld)\n", info.column[i].xmax,info.column[i].ymax);

       if(info.column[i].wcflg) 
       {
          fprintf(outfptr,"       world coord:  (%g,%g)\n",
             info.column[i].rxmax,info.column[i].rymax);
       }

       fprintf(outfptr,"              mean:  %.10g\n", info.column[i].mean);
       fprintf(outfptr,"            median:  %.10g\n", info.column[i].median);
       fprintf(outfptr,"             sigma:  %.10g\n", info.column[i].std);
       fprintf(outfptr,"               sum:  %.10g\n", info.column[i].sum);
       if(clip && !centroid) {
           if(info.column[i].converge) 
               fprintf(outfptr,"            cnvrgd:  YES\n");
           else
               fprintf(outfptr,"            cnvrgd:  NO\n");
       }

       if(centroid) {
           fprintf(outfptr,"      cntrd[pixel]:  (%g,%g)\n", info.column[i].cx,info.column[i].cy);

           if(info.column[i].wcflg) 
           {
              fprintf(outfptr,"      cntrd[world]:  (%g,%g)\n",
                info.column[i].rcx,info.column[i].rcy);
           }

           fprintf(outfptr,"sigma_cntrd[pixel]:  (%g,%g)\n",
             info.column[i].dx,info.column[i].dy);

           if(info.column[i].wcflg) 
           {
              fprintf(outfptr,"sigma_cntrd[world]:  (%g,%g)\n",
                info.column[i].rdx,info.column[i].rdy);
           }
       }

       fprintf(outfptr,"      #good pixels:  %ld\n", info.column[i].good);
       fprintf(outfptr,"      #null pixels:  %ld\n", info.column[i].null);
    } 
    else  /* statistics of a table column */
    {
       if( strcmp(info.column[i].unit," ") !=0 ) 
       {
          fprintf(outfptr,"%s[%s]:\n", info.column[i].name,info.column[i].unit);
       }
       else 
       {
          fprintf(outfptr,"%s:\n", info.column[i].name);
       }
       if(info.column[i].dtype < 0 )
       {
         fprintf(outfptr,"Not computing statistics for variable-length array column.\n");
       }
       if (info.column[i].dtype  == TSTRING || info.column[i].dtype  == TLOGICAL ) {
         continue;
	 
       } else if (info.column[i].dtype  == TFLOAT) {
           fprintf(outfptr,"   min:       %.6g\n", info.column[i].min);
           fprintf(outfptr,"   max:       %.6g\n", info.column[i].max);
           fprintf(outfptr,"  mean:       %.6g\n", info.column[i].mean);
           fprintf(outfptr,"median:       %.6g\n", info.column[i].median);
           fprintf(outfptr," sigma:       %.6g\n", info.column[i].std);
           fprintf(outfptr,"   sum:       %.6g\n", info.column[i].sum);

       } else {
           fprintf(outfptr,"   min:       %.15g\n", info.column[i].min);
           fprintf(outfptr,"   max:       %.15g\n", info.column[i].max);
           fprintf(outfptr,"  mean:       %.15g\n", info.column[i].mean);
           fprintf(outfptr,"median:       %.15g\n", info.column[i].median);
           fprintf(outfptr," sigma:       %.15g\n", info.column[i].std);
           fprintf(outfptr,"   sum:       %.15g\n", info.column[i].sum);
       }

       if(clip)
       {
           if(info.column[i].converge) 
               fprintf(outfptr,"cnvrgd:       YES\n");
           else
               fprintf(outfptr,"cnvrgd:       NO\n");
       }
       fprintf(outfptr,"  good:       %ld\n", info.column[i].good);
       fprintf(outfptr,"  null:       %ld\n", info.column[i].null);
    }
  } /* loop over all columns */

    return 0;
}

int ftstat_putpar(hduInfo * info, int centroid, int* status)
{
    int n;
    char msg[MAXMSG];

    if(*status) return *status;

    n =info->ncols -1;

    if((*status = PILPutReal("min",info->column[n].min))) 
    {
        sprintf(msg,"Error writing the 'min' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutReal("max",info->column[n].max))) 
    {
        sprintf(msg,"Error writing the 'max' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutInt("xmin",info->column[n].xmin))) 
    {
        sprintf(msg,"Error writing the 'xmin' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutInt("ymin",info->column[n].ymin))) 
    {
        sprintf(msg,"Error writing the 'ymin' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutInt("xmax",info->column[n].xmax))) 
    {
        sprintf(msg,"Error writing the 'xmax' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }
    if((*status = PILPutInt("ymax",info->column[n].ymax))) 
    {
        sprintf(msg,"Error writing the 'ymax' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutReal("mean",info->column[n].mean))) 
    {
        sprintf(msg,"Error writing the 'mean' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutReal("median",info->column[n].median))) 
    {
        sprintf(msg,"Error writing the 'median' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutReal("sigma",info->column[n].std))) 
    {
        sprintf(msg,"Error writing the 'sigma' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutReal("sum",info->column[n].sum))) 
    {
        sprintf(msg,"Error writing the 'sum' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutInt("null",info->column[n].null))) 
    {
        sprintf(msg,"Error writing the 'null' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutInt("good",info->column[n].good))) 
    {
        sprintf(msg,"Error writing the 'good' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }

    if((*status = PILPutReal("xcent",info->column[n].cx))) 
    {
        sprintf(msg,"Error writing the 'xcent' parameter");
        HD_ERROR_THROW(msg,*status);
        return *status;
    }
    if((*status = PILPutReal("ycent",info->column[n].cy))) 
    {
       sprintf(msg,"Error writing the 'ycent' parameter");
       HD_ERROR_THROW(msg,*status);
       return *status;
    }
    if((*status = PILPutReal("xsigma",info->column[n].dx))) 
    {
       sprintf(msg,"Error writing the 'xsigma' parameter");
       HD_ERROR_THROW(msg,*status);
       return *status;
    }
    if((*status = PILPutReal("ysigma",info->column[n].dy))) 
    {
       sprintf(msg,"Error writing the 'ysigma' parameter");
       HD_ERROR_THROW(msg,*status);
       return *status;
    }

    return *status;
}

void ftstat_free(hduInfo * info)
{
    int i;
    for (i=0; i<info->ncols; i++)
    {
       if(info->column[i].nulflag) free(info->column[i].nulflag);
       if(info->column[i].dvalue) free(info->column[i].dvalue);
    }
    free(info->column);
}

/*--------------------------------------------------------------------------*/

/*
 * This Quickselect routine is based on the algorithm described in
 * "Numerical recipes in C", Second Edition,
 * Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5
 * This code by Nicolas Devillard - 1998. Public domain.
 * http://ndevilla.free.fr/median/median/src/quickselect.c
 *
 * This routine selects & returns the median value of the array arr[].
 * If there are an even number of elements (n), the "lower median" is
 * selected.  The array arr[] is scrambled, so users should operate on
 * a scratch array if they wish the values to be preserved.
 */

/*--------------------------------------------------------------------------*/

#define ELEM_SWAP(a,b) { register double t=(a);(a)=(b);(b)=t; }

static double ftstat_median_double(double arr[], int n)
{
    int low, high ;
    int median;
    int middle, ll, hh;

    low = 0 ; high = n-1 ; median = (low + high) / 2;
    for (;;) {
        if (high <= low) /* One element only */
            return arr[median] ;

        if (high == low + 1) {  /* Two elements only */
            if (arr[low] > arr[high])
                ELEM_SWAP(arr[low], arr[high]) ;
            return arr[median] ;
        }

    /* Find median of low, middle and high items; swap into position low */
    middle = (low + high) / 2;
    if (arr[middle] > arr[high])    ELEM_SWAP(arr[middle], arr[high]) ;
    if (arr[low] > arr[high])       ELEM_SWAP(arr[low], arr[high]) ;
    if (arr[middle] > arr[low])     ELEM_SWAP(arr[middle], arr[low]) ;

    /* Swap low item (now in position middle) into position (low+1) */
    ELEM_SWAP(arr[middle], arr[low+1]) ;

    /* Nibble from each end towards middle, swapping items when stuck */
    ll = low + 1;
    hh = high;
    for (;;) {
        do ll++; while (arr[low] > arr[ll]) ;
        do hh--; while (arr[hh]  > arr[low]) ;

        if (hh < ll)
        break;

        ELEM_SWAP(arr[ll], arr[hh]) ;
    }

    /* Swap middle item (in position low) back into correct position */
    ELEM_SWAP(arr[low], arr[hh]) ;

    /* Re-set active partition */
    if (hh <= median)
        low = ll;
        if (hh >= median)
        high = hh - 1;
    }
}

#undef ELEM_SWAP
